/*
		JAS v1.00							  (c) 1997 NikSoft International
		--------------------------------------------------------------------
		KEYTBL.C

		Generates the hash table for the MYASM	reserved words, register
		names, instruction names and special keywords.

		It takes as input a file with the specification of all the
		keywords used and generates as output a C #include file with
		the static definition of the keywords hash table.
		The 'separate chain' method is used to cope with collisions.

		Monday, November 25, 1996	Version 1.00
		Tuesday, December 10, 1996
			Added a function to compute mean access cost.
			The load factor for the Hash table is now choosen via
			a parameter on the command line. If none specified a default
			value of 0.80 will be taken.
		Thursday, December 12, 1996
			Policy for table generation changed. The symbols list is scanned
			twice: the first scan is used to allocate home bucket for non
			colliding symbols, the second one is used to allocate the
			remaining symbols. This hack should reduce the mean length of a
			chain used to access a symbol. Symbols that collide because
			their home bucket is occupied by a symbol not belonging to that
			home bucket are now favoured.
		Thursday, February 20, 1997
			The original project MYASM is now called JAS the Jet ASsembler,
			I find this name more suitable and pretty! :)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

/* Length of the NULL terminated string for an identifier */
#define IDELEN						33
#define EXPECTED_LF					0.80

enum errors {
	OK_ERR = 0,
	IO_ERR,
	READ_ERR,
	MEM_ERR,
	SYN_ERR,
	OPT_ERR,
	MAXERR
};

enum tokens {
	TOKEN_EOF= 0,
	TOKEN_CLASS,
	TOKEN_ID,
	MAXTOKEN
};

typedef struct symbol {
	char *name;				/* Name of the symbol */
	char *Token;			/* Name of the token  */
	char *Info;				/* Name of the token info constant */
} SYMBOL;

typedef struct bucket {
	SYMBOL value;			/* Symbol contained into the bucket */
	struct bucket *Next;	/* Pointer to the next symbol in the list */
} BUCKET;

typedef struct twoptrs {
	BUCKET *Bck;
	struct twoptrs *Next;
} TWOPTRS;

/* -------- Static Datas ---------------------------------------------- */

static char DefSuffix[] = ".h";
static char *errmsg[] = {
	"No error",
	"Couldn't open file: %s.\n",
	"Error reading file: %s.\n",
	"Memory allocation error in %s().\n",
	"Syntax error at line %u.\n",
	"Invalid parameter: %s.\n"
};

static char TokenVal[IDELEN];

static BUCKET *BucketList;
static TWOPTRS *HashTable;

static int SymbolCount;
static int NoSlot;
static int Collision;
static int lineno;
static float ExpLF;

static char headMsg[]=
	"/* Source code automatically generated by keytbl, don't edit */\n\n";

static char TypeDecl[]=
	"typedef struct tblbucket {\n"
	"    char *sym_name;\n"
	"    int  Token;\n"
	"    int  Info;\n"
	"    struct tblbucket *Next;\n"
	"} TBLBUCKET;\n";


/* -------- Prototypes Of Local Functions ----------------------------- */

static void		error(int, ...);
static FILE		*OpenFile(char *, char *);
static int		GetToken(FILE *);
static int		ParseFile(FILE *);
static int		Hash(char *);
static int		GenerateTable(FILE *);
static BUCKET	*NewBucket(char *);
static float	AccessCost();

/* -------------------------------------------------------------------- */

void __crt0_load_environment_file(char *null)
{
}

char **__crt0_glob_function(char *null)
{
	return NULL;
}

int main(int argc, char *argv[])
{
	FILE	*Input, *Output;
	char	*OutputName, *temp;

	puts("(c) 1996 NikSoft International Keyword Table Generator");

	if ( argc == 1 ) {
		puts("Usage: keytbl [-load factor] <file-name> [<out-file>]");
		return -1;
	};

	if ( !(ExpLF= -atof(argv[1])) ) {
		if ( *argv[1] == '-' ) {
			error(OPT_ERR, argv[1]);
			return -1;
		};

		ExpLF= EXPECTED_LF;
	}
	else {
		if ( ExpLF > 1.0 ) {
			error(OPT_ERR, argv[1]);
			return -1;
		};

		/* Skip the Load factor parameter */
		argc--;
		argv++;
	}

	if ( argc > 2 ) {
		OutputName= (char *)malloc(strlen(argv[2]) + 5);
		strcpy(OutputName, argv[2]);
		if ( !strrchr(OutputName, '.') ) strcat(OutputName, DefSuffix);
	}
	else {
		OutputName= (char *)malloc(strlen(argv[1]) + 5);
		strcpy(OutputName, argv[1]);

		if ( (temp= strrchr(OutputName, '.')) ) *temp= '\0';
		strcat(OutputName, DefSuffix);
	}

	/* Opens the files */
	Input = OpenFile(argv[1], "r");
	Output= OpenFile(OutputName, "w");

	/* Initializes some variables */
	SymbolCount= 0;
	Collision  = 0;
	lineno	   = 1;

	switch ( ParseFile(Input) ) {
		case SYN_ERR:
			error(SYN_ERR, lineno);
			return -1;

		case -1:
			error(READ_ERR, argv[1]);
			return -1;
	};

	/* Now, generate the table */
	return GenerateTable(Output);
}

/* -------- Local Functions ------------------------------------------- */


/*
	Function......: error
	Description...: Error handling function.
	Parameters....:
					ErrNum				Error number.
					...					General info regarding the error.
*/
static void error(int ErrNum, ...)
{
	va_list arg_ptr;

	va_start(arg_ptr, ErrNum);

	if ( ErrNum >= MAXERR ) {
		fputs("Undefined Error", stderr);
	}
	else {
		vfprintf(stderr, errmsg[ErrNum], arg_ptr);
	};

	va_end(arg_ptr);
}


/*
	Function......: OpenFile
	Description...: It is a simple wrapper for the stdio library function
					fopen().
	Parameters....:
					Name			Name of the file to be opened.
					Mode			Open mode string.
*/
static FILE *OpenFile(char *Name, char *Mode)
{
	register FILE *RetVal;

	RetVal= fopen(Name, Mode);

	if ( RetVal == NULL ) {
		error(IO_ERR, Name);
		exit(-1);
	};

	return RetVal;
}


/*
	Function......: GetToken
	Description...: Reads and returns the next token in the file.
	Parameters....:
					Source				pointer to the scanned file.
*/
static int GetToken(FILE *Source)
{
	int ReadChar;

	/* Skips white spaces */
start:
	while ( (ReadChar = fgetc(Source)) == '\t' || ReadChar == ' ' );

	switch ( ReadChar ) {
		case '#':
		    /* Eat up a comment line */
		    while ( (ReadChar = fgetc(Source)) != -1 && ReadChar != '\n');

		    /* EOF or error encountered, gets out of the loop */
		    if ( ReadChar == -1 )
		case -1:
				return TOKEN_EOF;

		case '\n':
			lineno++;
			goto start;

		default:
			ungetc(ReadChar, Source);
			if ( fscanf(Source, "%s", TokenVal) != 1 ) return -1;
			if ( !strcmp(TokenVal, "token") ) return TOKEN_CLASS;
			return TOKEN_ID;
	};
}


/*
	Function......: ParseFile
	Description...: Parses the input file counting the symbols and storing
					them into the BucketList list.
					Returns zero if successfull, return an error code
					otherwise.
	Parameters....:
					Source			pointer to the file to be parsed.
*/
static int ParseFile(FILE *Source)
{
	register int  Tkn;
	register char *TokenClass= NULL;
	BUCKET	**NextInList= &BucketList;
	BUCKET	*CurrBucket;

	Tkn= GetToken(Source);

	while ( Tkn != -1 && Tkn != TOKEN_EOF ) {

		switch ( Tkn ) {
			case TOKEN_CLASS:
				/* Definition of a new class of keyword */
				if ( GetToken(Source) == TOKEN_ID ) {
					TokenClass= (char *)malloc(IDELEN);
					strcpy(TokenClass, TokenVal);
					Tkn= GetToken(Source);
				}
				else {
					return SYN_ERR;
				};
				break;

			case TOKEN_ID:
				/* Definition of a new keyword.
		           Allocates a new bucket */
		        CurrBucket= NewBucket(TokenClass);
		        CurrBucket->Next= NULL;
		        SymbolCount++;

		        /* Updates the list links */
		        *NextInList= CurrBucket;
		        NextInList = &CurrBucket->Next;

				strcpy(CurrBucket->value.name, TokenVal);

				/* Optional Info value*/
				if ( (Tkn= GetToken(Source)) == TOKEN_ID ) {
					strcpy(CurrBucket->value.Info, TokenVal);
					Tkn= GetToken(Source);
				}
				else {
					strcpy(CurrBucket->value.Info, "-1");
				};
				break;

			default:
				return SYN_ERR;
		};
	}

	fclose(Source);
	return ( Tkn == TOKEN_EOF ? 0 : -1);
}


/*
	Function......: NewBucket
	Description...: Allocates a new bucket.
					Returns the pointer to the new allocated bucket if
					successful, exits the program otherwise.
	Parameters....:
					Class			pointer to the name of the keyword
									class, i.e. the token value that will be
									returned by the parser.
*/
static BUCKET *NewBucket(char *Class)
{
	register char *RetVal;

	/* Allocates in one shot all the memory necessary to a bucket */
	RetVal= (char *)malloc(sizeof(BUCKET) + 2*IDELEN);

	if ( !RetVal ) {
		error(MEM_ERR, "NewBucket");
		exit(-1);
	};

	/* Now adjusts the pointers */
	((BUCKET *)RetVal)->value.name = RetVal + sizeof(BUCKET);
	((BUCKET *)RetVal)->value.Info = RetVal + sizeof(BUCKET) + IDELEN;
	((BUCKET *)RetVal)->value.Token= Class;

	return (BUCKET *)RetVal;
}


/*
	Function......: Hash
	Description...: Hashes a string.
					Returns the hash value for the string.
	Parameters....:
					Name			string to be hashed.
*/
static int Hash(char *Name)
{
	register unsigned sum= 0;

	while ( *Name ) {
		sum += *Name;
		Name++;
	};

	return (sum % NoSlot);
}


/*
	Function......: GenerateTable
	Description...: Generates the hash table for the keywords and
					writes it in a standard C #include file.
	Parameters....:
					Output				Pointer to the FILE struct
										for the file to be genarated.
*/
static int GenerateTable(FILE *Output)
{
	register BUCKET *BckCrsr;
	register TWOPTRS *TblCrsr;
	register TWOPTRS *LastBck;
	static char pointerStr[64];
	int i;

	fputs(headMsg, Output);
	fputs(TypeDecl, Output);

	fputs("\nstatic TBLBUCKET HashTable[]= {\n", Output);

	NoSlot= (int)((float)SymbolCount / ExpLF);
	HashTable= (TWOPTRS *)malloc(sizeof(TWOPTRS) * NoSlot);

	if ( !HashTable ) {
		error(MEM_ERR, "GenerateTable");
		return -1;
	};

	/* Clears the array */
	memset(HashTable, 0, sizeof(TWOPTRS) * NoSlot);

	/* Loads the table */

	/* First scan */
	for ( BckCrsr= BucketList; BckCrsr; BckCrsr= BckCrsr->Next) {
		/* Calculates the hash value of the symbol name */
		TblCrsr= &HashTable[ Hash(BckCrsr->value.name) ];

		if ( !TblCrsr->Bck ) {
			/* Occupie the bucket */
		    TblCrsr->Bck = BckCrsr;
		    TblCrsr->Next= NULL;
		}
	};

	/* Second scan */
	for ( BckCrsr= BucketList; BckCrsr; BckCrsr= BckCrsr->Next) {
		/* Calculates the hash value of the symbol name */
		TblCrsr= &HashTable[ Hash(BckCrsr->value.name) ];

		if ( TblCrsr->Bck != BckCrsr ) {
			/* Registers the collision */
			Collision++;

			/* Follows the chain of colliding bucket */
			while ( TblCrsr->Next ) {
				TblCrsr= TblCrsr->Next;
			};

			/* LastBck points to the end of the chain */
			LastBck= TblCrsr;

			/* Scans the array of bucket looking for a free bucket */
			while ( TblCrsr->Bck ) {
				TblCrsr++;

				if ( TblCrsr == &HashTable[NoSlot]) {
					TblCrsr= HashTable;
				};
			};

			LastBck->Next= TblCrsr;
		    TblCrsr->Bck = BckCrsr;
		    TblCrsr->Next= NULL;
		};
	};

	/* Writes the table on the file */
	for ( i= 0, TblCrsr= HashTable; i < NoSlot;
		  i++, TblCrsr++) {

		if ( (BckCrsr= TblCrsr->Bck) ) {

			if ( TblCrsr->Next ) {
				sprintf(pointerStr, "&HashTable[%u]",
					(TblCrsr->Next - HashTable));
			}
			else {
				strcpy(pointerStr, "(TBLBUCKET *)0");
			}

			fprintf(Output, "/* %03u */{ \"%s\", %s, %s, %s },\n",
				i,
				BckCrsr->value.name,
				BckCrsr->value.Token,
				BckCrsr->value.Info,
				pointerStr);
		}
		else {
			fprintf(Output,
				"/* %03u */{ (char *)0, 0, 0, (TBLBUCKET *)0 },\n", i);
		}
	};

	fputs("};", Output);

	printf("Total symbols......: %u\n", SymbolCount);
	printf("Total slots........: %u\n", NoSlot);
	printf("Collisions detected: %u\n", Collision);
	printf("Mean access cost...: %f\n", AccessCost());

	return fclose(Output);
}


/*
	Function......: AccessCost
	Description...: Computes the Hash table mean access cost by
					accessing every symbol in the Symbol list, counting
					the number of memory access for each symbol and then
					it takes the mean of this figure.
	Parameters....: None.
*/
static float AccessCost()
{
	unsigned Accesses= 0;
	register BUCKET *SymCrsr= BucketList;
	register TWOPTRS *TblCrsr;

	while ( SymCrsr ) {
		TblCrsr= &HashTable[ Hash(SymCrsr->value.name) ];
		Accesses++;

		while ( TblCrsr->Bck != SymCrsr ) {
			Accesses++;
			TblCrsr= TblCrsr->Next;
		};

		SymCrsr= SymCrsr->Next;
	};

	return ((float)Accesses / SymbolCount);
}

/* -------- End Of File ----------------------------------------------- */
