/*
		JAS v1.00							  (c) 1996 NikSoft International
		--------------------------------------------------------------------
		SYMMNGR.C

		Symbol table management module.

		06/11/96 19.56	Version 1.00

		Tuesday, November 26, 1996
			All the reserved keywords (register names, istructions an so on)
			are now collected in an hash table automatically generated from
			a keywords specification source file parsed by the KEYTBL
			program. Only the conditional istructions are still handled
			in a separate table, while the conditions are stored in the hash
			table.
*/

#include <ctype.h>

#include "globals.h"
#include "jaslib.h"

#include "ptypes.h"
#include "symmngr.h"
#include "parser.h"
#include "keywords.h"           /* This file contains the declaration
								   of the keyword hash table */

#define DICTSIZE				211

extern int TokenInfo;

typedef struct stbucket {
	SYMBOL value;               /* Symbol stored in the bucket */

	int	   BckNo;				/* Bucket to which the symbol belongs */
	struct stbucket *Next;	    /* Pointers for the overflow handling */
	struct stbucket *Prev;
} STBUCKET;

/* -------- Static datas ---------------------------------------------- */

static STBUCKET	*dictionary[DICTSIZE];

static char *icn[] = {		       /* conditional instructions */
	"CMOV", "J", "SET"
};

static int ico[] = {		       /* and the corresponding opcodes */
	I_CMOVcc, I_Jcc, I_SETcc
};

/* -------- Prototypes of static functions ---------------------------- */

static int	HashSearch(char *, int);
static int	Hash(char *, int, int);
static void FreeBucket(STBUCKET *);

/* -------- Public Functions ------------------------------------------ */

/*
	Function......: LookUp
	Description...: This function searches the specified identifier
					in the symbol table. If the symbol is present it
					returns the pointer to the SYMBOL struct for the
					specified identifier, otherwise it allocates a new
					SYMBOL struct and returns the pointer to it.
	Parameters....:
					IdName				Identifier to be looked up
					Case				If zero ignores the case
										of the identifier.
*/
SYMBOL *LookUp(char *IdName, int Case)
{
	register STBUCKET *SymBucket;
	register STBUCKET *LastBucket= NULL;
	register int BckNo;

	/* Get the correct bucket */
	BckNo= Hash(IdName, DICTSIZE, Case);
	SymBucket= dictionary[BckNo];

	/* Follow the bucket chain */
	while ( SymBucket && strcmp(SymBucket->value.Name, IdName) ) {
		LastBucket= SymBucket;
		SymBucket = SymBucket->Next;
	};

	if ( !SymBucket ) {
		/* Symbol not present, allocate a new bucket and link
		   it to the chain */
		SymBucket= (STBUCKET *)LIB_malloc(sizeof(STBUCKET));

		bzero(&SymBucket->value, sizeof(SYMBOL));
		SymBucket->value.Name     = LIB_strdup(IdName);
		SymBucket->value.Type     = NO_TYPE;
		SymBucket->value.ExtraLong= -1;
		SymBucket->value.ExtraPtr = NULL;

		SymBucket->BckNo= BckNo;
		SymBucket->Next = NULL;
		SymBucket->Prev = LastBucket;

		if ( !LastBucket ) {
			dictionary[BckNo]= SymBucket;
		} else {
			LastBucket->Next = SymBucket;
		}
	};

	/* Got him! */
	return &SymBucket->value;
}


/*
	Function......: IsKeyword
	Description...:	Check if the passed string is a reserved keyword.
					Returns a valid token for the keyword and updates
					the yylval and TokenInfo variable if successfull,
					returns -1 otherwise.
	Parameters....:
					name			string to be analyzed
*/
int IsKeyword(char *name)
{
	register int anInt;

	/* Searches for the specified keyword in the hash table */
	anInt     = HashSearch(name, 0);
	TokenInfo = -1;

	if ( anInt > 0 ) return anInt;

	/* Look if it is a conditional instruction */
	for ( anInt= 0; anInt < elements(icn); anInt++ ) {

		if ( LIB_strnicmp(icn[anInt], name, strlen(icn[anInt])) == 0 ) {
		/* Ok, it's a conditional istruction, but which one?
		   Look for the condition code */
		    HashSearch( &name[strlen(icn[anInt])], 0);
			if ( (TokenInfo= yylval.intval) < 0 ) return -1;

			/* Restores the istruction opcode */
			yylval.intval= ico[anInt];
			return TOKEN_INSN;
		};
	}

	/* The parameter is not a reserved keyword... return */
	return -1;
}


/*
	Function......: FreeSymbolList
	Description...: Frees the memory allocated for the list of identifier
					connected to a specified identifier.
					The Head of the list is not deleted by the function.
	Parameters....:
					ListHead			Pointer to the head of the list.
*/
void FreeSymbolList(SYMBOL *ListHead)
{
	register SYMBOL *ListCrsr;
	register SYMBOL *Next;

	if ( ListHead == NULL ) return;

	ListCrsr= ListHead->NextLink;
	while ( ListCrsr ) {
		Next= ListCrsr->NextLink;
		FreeBucket((STBUCKET *)ListCrsr);
		ListCrsr= Next;
	};

	ListHead->NextLink= NULL;
}


/*
	Function......: FreeSymbol
	Description...: Removes a symbol and all the symbols linked to it from
					the symbol table, freeing up the memory allocated for
					them.
	Parameters....:
					SymItem				Symbol to be removed.
*/
void FreeSymbol(SYMBOL *SymItem)
{
	if ( SymItem == NULL ) return;
	FreeSymbolList(SymItem);
	FreeBucket((STBUCKET *)SymItem);
}

/*
	Function......: LinkSymbol
	Description...: Adds a symbol at the head of the symbol list connected
					to the specified identifier.
	Parameters....:
					ListHead			Symbol to which the new symbol has
										to be linked to.
					SymItem				Symbol to be added to the list.
*/
void LinkSymbol(SYMBOL *ListHead, SYMBOL *SymItem)
{
	register SYMBOL *List;

	if ( !ListHead || !SymItem ) return;

	List= ListHead->NextLink;
	ListHead->NextLink= SymItem;
	SymItem->Head= ListHead;

	if ( List ) SymItem->NextLink= List;
}


/* -------- Local Functions ------------------------------------------- */

/*
	Function......: Hash
	Description...: Implements a simple hash function.
					The characters of the identifier are summed and the
					result is divided modulo the number of available
					buckets. The result of this division is then returned.
					NOTE: This function is used to access every Hash table
						  of the module.
	Parameters....:
					id			identifier to be hashed.
					Buckets		Available bucket.
					Case		if zero ignores the id case.
*/
static int Hash(register char *id, int Buckets, int Case)
{
	register int sum= 0;

	while ( *id ) {
		sum += ( Case ? *id : toupper(*id) );
		id++;
	};

	return (sum % Buckets);
}


/*
	Function......: HashSearch
	Description...: Searches for the specified string in the reserved
					words hash table.
					Returns the token value for the keyword and updates
					the global variable yylval if successfull, -1 otherwise.
	Parameters....:
					name				pointer to the string to be
										searched.
					Case				If zero convert ignores the
										string case.
*/
static int HashSearch(char *name, int Case)
{
	register TBLBUCKET *TblCrsr;
	register int (*CmpFunc)(const char *, const char *);

	yylval.intval= -1;
	CmpFunc= ( Case ? strcmp : LIB_stricmp );

	/* Hash the identifier and take the bucket address */
	TblCrsr= &HashTable[ Hash(name, elements(HashTable), Case) ];

	/* We have hit an empty bucket so the identifier is not present. */
	if ( TblCrsr->sym_name == NULL) {
		return -1;
	}

	/* Follow the chain looking for the parameter */
	while ( TblCrsr && CmpFunc(TblCrsr->sym_name, name) ) {
		TblCrsr= TblCrsr->Next;
	};

	/* We are at the end of the chain, the parameter is not in the
	   hash table */
	if ( !TblCrsr ) return -1;

	/* We got the name! Updates the global variable and returns the
	   keyword token */
	yylval.intval= TblCrsr->Info;
	return TblCrsr->Token;
}


/*
	Function......: FreeBucket
	Description...: Frees a bucket.
	Parameters....:
					Bck				pointer to the bucket to be freed.
*/
static void FreeBucket(STBUCKET *Bck)
{
	/* Unlink the bucket from the overflow chain ...
	   ... first the Prev pointer ... */
	if ( Bck->Prev ) {
	    Bck->Prev->Next= Bck->Next;
	} else {
		dictionary[Bck->BckNo]= Bck->Next;
	};

	/* ... then the Next pointer */
	if ( Bck->Next ) {
		Bck->Next->Prev= Bck->Prev;
	};

	/* Free the memory */
	LIB_free(Bck);
}

/* -------- End Of File ----------------------------------------------- */
